module sorter
	contains
		! Seeds the built in random number generator with the date, ensuring a different sequence on each execution.
		subroutine randomize()
			integer, dimension(:), pointer :: date
			integer :: seedSize
	
			call random_seed(size=seedSize)
			if(seedSize < 8) seedSize = 8
			allocate(date(seedSize))
			date = 0
			call date_and_time(values=date)
			call random_seed(put=date)
		end subroutine

		! Do a simple insertion sort. Has quadratic complexity, but the inner loop is short.
		subroutine insertionSort(array)
			real, dimension(:), intent(inout) :: array
			integer :: i, j
			real :: temp
			do i = lbound(array, 1), ubound(array, 1)
				temp = array(i)
				do j = i - 1, 1, -1
					if(array(j) <= temp) then
						array(j + 1) = temp
						exit
					else
						array(j + 1) = array(j)
					end if
				end do
			end do
		end subroutine
		
		! Do a simple bucket sort. Has nlogn complexity, but with a rather long inner loop.
		recursive subroutine bucketSort(array, lowerBound, upperBound)
			real, dimension(:), intent(inout) :: array
			real, intent(in) :: lowerBound, upperBound	! Should be the min/max of the values to be sorted.
			real :: buffer(lbound(array, 1):ubound(array, 1))
			integer, parameter :: bucketCount = 10
			integer :: bucketSizes(0:bucketCount), bucketIndices(0:bucketCount), i, curBucket, curSize, nextBucketStart
			
			! Determine the size of each bucket.
			bucketSizes = 0
			do i = lbound(array, 1), ubound(array, 1)
				curBucket = int(bucketCount*(array(i) - lowerBound)/(upperBound - lowerBound))
				bucketSizes(curBucket) = bucketSizes(curBucket) + 1
			end do
			! Determine the start of each bucket.
			i = 1
			do curBucket = 0, bucketCount
				bucketIndices(curBucket) = i
				i = i + bucketSizes(curBucket)
			end do
			! Sort the values into the buckets.
			do i = lbound(array, 1), ubound(array, 1)
				curBucket = int(bucketCount*(array(i) - lowerBound)/(upperBound - lowerBound))
				buffer(bucketIndices(curBucket)) = array(i)
				bucketIndices(curBucket) = bucketIndices(curBucket) + 1
			end do
			! Sort the buckets, bucketIndices now point past the end of the bucket.
			do curBucket = 0, bucketCount
				curSize = bucketSizes(curBucket)
				nextBucketStart = bucketIndices(curBucket)
				if(curSize > 0) then
					call bucketSort(buffer(nextBucketStart - curSize : nextBucketStart - 1), &
						& lowerBound + curBucket*(upperBound - lowerBound)/bucketCount, &
						& lowerBound + (curBucket + 1)*(upperBound - lowerBound)/bucketCount)
				end if
			end do
			! Copy back.
			array = buffer
		end subroutine
		
		subroutine startBucketSort(array)
			real, intent(inout) :: array(:)
			call bucketSort(array, minval(array), maxval(array))
		end subroutine
end module

program r2d2
	use sorter
	real :: someData(1:20)
	call randomize()

	print*,"insertion sort:"
	call random_number(someData)
	call insertionSort(someData)
	do i = lbound(someData, 1), ubound(someData, 1)
		print*, someData(i)
	end do

	print*
	print*,"bucket sort:"
	call random_number(someData)
	call startBucketSort(someData)
	do i = lbound(someData, 1), ubound(someData, 1)
		print*, someData(i)
	end do
end program
